import unittest

from pyspark.sql import SparkSession
import pyspark.sql.functions as fn
from pyspark.sql.types import StringType

from trend_etl.test_utils import get_kafka_df, getspark, get_df
import trend_etl.schema.fields as fields
import trend_etl.schema.canonical as canonical
import trend_etl.functions.ingest_versioning as ingest
from .parsing import (
    parse_subreddits,
    parse_reddit_submissions,
    parse_reddit_replies,
    parse_facebook_groups,
    parse_facebook_posts,
    parse_facebook_comments,
    parse_guilds,
    parse_channels,
    parse_members,
    parse_messages,
    parse_reactions,
    normalize_reddit,
)


class TestFunctions(unittest.TestCase):
    spark: SparkSession

    @classmethod
    def setUpClass(cls) -> None:
        cls.spark = getspark()

    def test_direct_to_binary_null_empty_string(self) -> None:
        input_df = self.__class__.spark.createDataFrame(
            [(None,)], schema="fruit string"
        )

        got_df = input_df.withColumn(
            "ingest_version", fn.md5(ingest.direct_to_binary(fn.col("fruit")))
        )

        # $ echo -n -e "" | md5
        # d41d8cd98f00b204e9800998ecf8427e

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "d41d8cd98f00b204e9800998ecf8427e",
            "doc version mismatch",
        )

        input_df = self.__class__.spark.createDataFrame([("",)], schema="fruit string")

        got_df = input_df.withColumn(
            "ingest_version", fn.md5(ingest.direct_to_binary(fn.col("fruit")))
        )

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "d41d8cd98f00b204e9800998ecf8427e",
            "doc version mismatch",
        )

    def test_direct_to_binary(self) -> None:
        data = [("apple",)]
        input_df = self.__class__.spark.createDataFrame(data, schema="fruit string")

        got_df = input_df.withColumn(
            "ingest_version", fn.md5(ingest.direct_to_binary(fn.col("fruit")))
        )

        # $ echo -n -e "\x61\x70\x70\x6C\x65" | md5
        # 1f3870be274f6c49b3e31a0c6728957f

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "1f3870be274f6c49b3e31a0c6728957f",
            "doc version mismatch",
        )

    def test_boolean_to_binary(self) -> None:
        data = [(True,)]
        input_df = self.__class__.spark.createDataFrame(data, schema="flag boolean")

        got_df = input_df.withColumn(
            "ingest_version", fn.md5(ingest.boolean_to_binary(fn.col("flag")))
        )

        # $ echo -n -e "\x00\x00\x00\x01" | md5
        # f1450306517624a57eafbbf8ed995985

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "f1450306517624a57eafbbf8ed995985",
            "doc version mismatch",
        )

    def test_array_to_binary_array_strings(self) -> None:
        data = [(["apple"],)]
        input_df = self.__class__.spark.createDataFrame(
            data, schema="fruits array<string>"
        )

        got_df = input_df.withColumn(
            "ingest_version",
            fn.md5(ingest.array_to_binary(ingest.direct_to_binary, fn.col("fruits"))),
        )

        # $ echo -n -e "\x61\x70\x70\x6C\x65" | md5
        # 1f3870be274f6c49b3e31a0c6728957f

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "1f3870be274f6c49b3e31a0c6728957f",
            "doc version mismatch",
        )

    def test_array_to_binary_array_structs(self) -> None:
        data = [
            (
                [
                    ("apple", 1, 42),  # Tuple representing the struct
                ],
            ),
        ]
        input_df = self.__class__.spark.createDataFrame(
            data,
            schema="rows array<struct<fruit string, picked_on int, number_collected int>>",
        )

        got_df = input_df.withColumn(
            "ingest_version",
            fn.md5(
                ingest.array_to_binary(
                    lambda elem: fn.concat(
                        ingest.direct_to_binary(elem.getField("fruit")),
                        ingest.direct_to_binary(elem.getField("picked_on")),
                        ingest.direct_to_binary(elem.getField("number_collected")),
                    ),
                    fn.col("rows"),
                )
            ),
        )

        # $ echo -n -e "\x61\x70\x70\x6C\x65\x00\x00\x00\x01\x00\x00\x00\x2A" | md5
        # 7abbf5554a6355f2572943f1aae9a338

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "7abbf5554a6355f2572943f1aae9a338",
            "doc version mismatch",
        )

    def test_direct_to_binary_concat(self) -> None:
        data = [("apple", 1, 42)]
        input_df = self.__class__.spark.createDataFrame(
            data, schema="fruit string, picked_on int, number_collected int"
        )

        got_df = input_df.withColumn(
            "ingest_version",
            fn.md5(
                fn.concat(
                    ingest.direct_to_binary(fn.col("fruit")),
                    ingest.direct_to_binary(fn.col("picked_on")),
                    ingest.direct_to_binary(fn.col("number_collected")),
                )
            ),
        )

        # $ echo -n -e "\x61\x70\x70\x6C\x65\x00\x00\x00\x01\x00\x00\x00\x2A" | md5
        # 7abbf5554a6355f2572943f1aae9a338

        self.assertEqual(got_df.count(), 1, "count mismatch")
        self.assertEqual(
            got_df.collect()[0]["ingest_version"],
            "7abbf5554a6355f2572943f1aae9a338",
            "doc version mismatch",
        )

    def test_subreddit_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "subreddits.jsonl"
        )

        parsed_df = input_df.select(
            parse_subreddits(
                fn.col(fields.bronze_value_field).cast(StringType())
            ).alias("json")
        )

        filtered = parsed_df.where(fn.col("json.kind") == fn.lit("t5"))

        got_df = filtered.select(
            ingest.ingest_version(
                normalize_reddit(fn.col("json").getField("data")),
                canonical.subreddit_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "subreddit_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_reddit_submission_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "submissions.jsonl"
        )

        parsed_df = input_df.select(
            parse_reddit_submissions(
                fn.col(fields.bronze_value_field).cast(StringType())
            ).alias("json")
        )

        got_df = parsed_df.select(
            ingest.ingest_version(
                normalize_reddit(fn.col("json").getField("data")),
                canonical.reddit_submission_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "submission_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_reddit_reply_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "replies.jsonl"
        )

        parsed_df = input_df.select(
            parse_reddit_replies(
                fn.col(fields.bronze_value_field).cast(StringType())
            ).alias("json")
        )

        got_df = parsed_df.select(
            ingest.ingest_version(
                normalize_reddit(fn.col("json").getField("data")),
                canonical.reddit_reply_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "reply_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_group_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "groups.jsonl"
        )

        parsed_df = input_df.select(
            parse_facebook_groups(
                fn.col(fields.bronze_value_field).cast(StringType())
            ).alias("json")
        )

        got_df = parsed_df.select(
            ingest.ingest_version(
                fn.col("json"),
                canonical.facebook_group_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "group_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_facebook_posts_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "posts.jsonl"
        )

        parsed_df = input_df.select(
            parse_facebook_posts(
                fn.col(fields.bronze_value_field).cast(StringType())
            ).alias("json")
        )

        got_df = parsed_df.select(
            ingest.ingest_version(
                fn.col("json"),
                canonical.facebook_post_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "post_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_facebook_comments_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "comments.jsonl"
        )

        parsed_df = input_df.select(
            parse_facebook_comments(
                fn.col(fields.bronze_value_field).cast(StringType())
            ).alias("json")
        )

        got_df = parsed_df.select(
            ingest.ingest_version(
                fn.col("json"),
                canonical.facebook_comment_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "comment_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_discord_guild_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "guilds.jsonl"
        )

        parsed_df = input_df.select(
            parse_guilds(fn.col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        got_df = parsed_df.select(
            ingest.ingest_version(
                fn.col("json"),
                canonical.discord_guild_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "guild_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_channel_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "channels.jsonl"
        )

        parsed_df = input_df.select(
            parse_channels(fn.col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        got_df = parsed_df.select(
            ingest.ingest_version(
                fn.col("json"),
                canonical.discord_channel_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "channel_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_discord_message_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "messages.jsonl"
        )

        parsed_df = input_df.select(
            parse_messages(fn.col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        got_df = parsed_df.select(
            ingest.ingest_version(
                fn.col("json"),
                canonical.discord_message_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "message_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_discord_text_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "messages.jsonl"
        )

        parsed_df = input_df.select(
            parse_messages(fn.col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        got_df = parsed_df.select(
            ingest.ingest_version(
                fn.col("json"),
                canonical.discord_text_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "text_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_discord_mention_user_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "messages.jsonl"
        )

        parsed_df = input_df.select(
            parse_messages(fn.col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        got_df = (
            parsed_df.select(
                fn.col("json.id").alias("message_id"),
                fn.explode("json.mentions").alias("mentions"),
            )
            .withColumn("user_id", fn.col("mentions").getField("id"))
            .drop("mentions")
            .select(
                ingest.ingest_version(
                    fn.struct(fn.col("*")), canonical.discord_mention_user_canonical
                ).alias("ingest_version")
            )
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "mention_user_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_discord_mention_channel_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "messages.jsonl"
        )

        parsed_df = input_df.select(
            parse_messages(fn.col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        got_df = (
            parsed_df.select(
                fn.col("json.id").alias("message_id"),
                fn.explode("json.mention_channels").alias("mention_channels"),
            )
            .withColumn("channel_id", fn.col("mention_channels").getField("id"))
            .drop("mention_channels")
            .select(
                ingest.ingest_version(
                    fn.struct(fn.col("*")), canonical.discord_mention_channel_canonical
                ).alias("ingest_version")
            )
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "mention_channel_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_discord_reaction_ingest_version(self) -> None:
        input_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "reactions.jsonl"
        )

        parsed_df = input_df.select(
            parse_reactions(fn.col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        got_df = parsed_df.select(
            ingest.ingest_version(
                fn.col("json")
                .withField("user_id", fn.col("json.user.id"))
                .dropFields("user"),
                canonical.discord_reaction_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "reaction_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_discord_member_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "members.jsonl"
        )

        parsed_df = source_df.select(
            parse_members(fn.col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        got_df = parsed_df.select(
            ingest.ingest_version(
                fn.col("json")
                .withField("user_id", fn.col("json.user.id"))
                .dropFields("user"),
                canonical.discord_member_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "member_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())

    def test_discord_user_ingest_version(self) -> None:
        source_df = get_kafka_df(
            self.__class__.spark, "test_data", "duplicated", "members.jsonl"
        )

        parsed_df = source_df.select(
            parse_members(fn.col(fields.bronze_value_field).cast(StringType())).alias(
                "json"
            )
        )

        got_df = parsed_df.select(
            ingest.ingest_version(
                fn.col("json.user"),
                canonical.discord_user_canonical,
            ).alias("ingest_version")
        )

        want_df = get_df(
            self.__class__.spark,
            got_df.schema,
            "test_data",
            "duplicated",
            "user_versions.jsonl",
        )

        self.assertEqual(want_df.collect(), got_df.collect())
