from typing import Optional
from datetime import datetime
import struct
import json

import pyspark.sql.functions as fn
from pyspark.sql.types import (
    TimestampType,
    BooleanType,
    BinaryType,
)


# arguments
#   date_as_str: date as a string in one of the expected formats, could be null
# return: column type or exception when string is not in the expected format
@fn.udf(returnType=TimestampType())
def conv_timestamp(date_as_str: Optional[str]) -> Optional[datetime]:
    if date_as_str is None or len(date_as_str) <= 0:
        return None

    try:
        return datetime.strptime(date_as_str, "%Y-%m-%dT%H:%M:%S.%f%z")
    except ValueError:
        pass

    return datetime.strptime(date_as_str, "%Y-%m-%dT%H:%M:%S%z")


@fn.udf(returnType=BooleanType())
def is_scientific(s: str) -> bool:
    return "E" in s or "e" in s


@fn.udf(returnType=BinaryType())
def double_to_binary(d: Optional[float]) -> bytes:
    if d is None:
        return b""
    return struct.pack(">d", d)


@fn.udf(returnType=TimestampType())
def discord_snowflake_timestamp(snowflake_id: str) -> datetime:
    # Discord epoch is January 1, 2015, at 00:00:00 UTC
    DISCORD_EPOCH = 1420070400000

    # Extract the timestamp from the snowflake_id
    timestamp = (int(snowflake_id) >> 22) + DISCORD_EPOCH

    # Convert the timestamp to seconds
    posix_timestamp = timestamp / 1000.0

    # Convert the timestamp to a datetime object
    date = datetime.fromtimestamp(posix_timestamp)

    return date


@fn.udf(returnType=BooleanType())
def annotation_matches_text(text: Optional[str], annontation_json: str) -> bool:
    def check_matches(text: str, entities_json: str) -> bool:
        entities = json.loads(entities_json)

        for entity in entities:
            start_index = entity["start_char"]
            end_index = entity["end_char"]
            span_text = "".join(list(text)[start_index:end_index])

            if span_text != entity["text"]:
                matches = False
                break
        else:
            matches = True

        return matches

    annontaion = json.loads(annontation_json)

    if annontaion["number_of_entities"] <= 0:
        return True

    if text is None:
        return False

    return check_matches(text, annontaion["entities"])
