import argparse
from typing import cast

from delta.tables import DeltaTable
from pyspark.sql import SparkSession, DataFrame
import pyspark.sql.functions as fn
from pyspark.sql.streaming import StreamingQuery

from trend_etl.functions.delta import (
    full_table_name,
    most_recent_version,
    earliest_version,
    merge_most_recent_version,
    merge_earliest_version,
)
import trend_etl.schema.args as arguments

program = "canonicalize_discord"


def write_guilds(
    spark: SparkSession,
    silver_catalog: str,
    target_catalog: str,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    silver_table = full_table_name(
        silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_discord_guilds_ao],
    )
    streaming = spark.readStream.format("delta").table(silver_table)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/"
        + arguments.schemas[arguments.schemas_gold]
        + "/"
        + "discord_guilds"
    )

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(
            spark,
            full_table_name(
                target_catalog,
                arguments.schemas[arguments.schemas_gold],
                "discord_guilds",
            ),
        )

        canonical = most_recent_version(batch_df, fn.col("source_keys.id"))

        return merge_most_recent_version(
            delta_table=delta_table, updates=canonical, merge_keys=["source_keys.id"]
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_channels(
    spark: SparkSession,
    silver_catalog: str,
    target_catalog: str,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    silver_table = full_table_name(
        silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_discord_channels_ao],
    )
    streaming = spark.readStream.format("delta").table(silver_table)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/"
        + arguments.schemas[arguments.schemas_gold]
        + "/"
        + "discord_channels"
    )

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(
            spark,
            full_table_name(
                target_catalog,
                arguments.schemas[arguments.schemas_gold],
                "discord_channels",
            ),
        )

        canonical = most_recent_version(batch_df, fn.col("source_keys.id"))

        return merge_most_recent_version(
            delta_table=delta_table, updates=canonical, merge_keys=["source_keys.id"]
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


# NOTE: need to materialize threads since we need the earliest record version
def write_threads(
    spark: SparkSession,
    silver_catalog: str,
    target_catalog: str,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    silver_table = full_table_name(
        silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_discord_channels_ao],
    )
    streaming = spark.readStream.format("delta").table(silver_table)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/"
        + arguments.schemas[arguments.schemas_gold]
        + "/"
        + "discord_threads"
    )

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(
            spark,
            full_table_name(
                target_catalog,
                arguments.schemas[arguments.schemas_gold],
                "discord_threads",
            ),
        )

        canonical = earliest_version(batch_df, fn.col("source_keys.id"))

        return merge_earliest_version(
            delta_table=delta_table, updates=canonical, merge_keys=["source_keys.id"]
        )

    return (
        # filter out non-threads channel records
        streaming.where(fn.col("type").isin(arguments.discord_thread_types))
        .writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_members(
    spark: SparkSession,
    silver_catalog: str,
    target_catalog: str,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    silver_table = full_table_name(
        silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_discord_members_ao],
    )
    streaming = spark.readStream.format("delta").table(silver_table)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/"
        + arguments.schemas[arguments.schemas_gold]
        + "/"
        + "discord_members"
    )

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(
            spark,
            full_table_name(
                target_catalog,
                arguments.schemas[arguments.schemas_gold],
                "discord_members",
            ),
        )

        canonical = most_recent_version(batch_df, fn.col("source_keys"))

        return merge_most_recent_version(
            delta_table=delta_table,
            updates=canonical,
            merge_keys=["source_keys.guild_id", "source_keys.user_id"],
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_messages(
    spark: SparkSession,
    silver_catalog: str,
    target_catalog: str,
    texts: DataFrame,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    silver_table = full_table_name(
        silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_discord_messages_ao],
    )
    streaming = spark.readStream.format("delta").table(silver_table)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/"
        + arguments.schemas[arguments.schemas_gold]
        + "/"
        + "discord_messages"
    )

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(
            spark,
            full_table_name(
                target_catalog,
                arguments.schemas[arguments.schemas_gold],
                "discord_messages",
            ),
        )

        canonical = earliest_version(batch_df, fn.col("source_keys.id"))

        canonical_with_text = canonical.join(texts, "text_version", "left_outer")

        return merge_earliest_version(
            delta_table=delta_table,
            updates=canonical_with_text,
            merge_keys=["source_keys.id"],
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_users(
    spark: SparkSession,
    silver_catalog: str,
    target_catalog: str,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    silver_table = full_table_name(
        silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_discord_users_ao],
    )
    streaming = spark.readStream.format("delta").table(silver_table)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/"
        + arguments.schemas[arguments.schemas_gold]
        + "/"
        + "discord_users"
    )

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(
            spark,
            full_table_name(
                target_catalog,
                arguments.schemas[arguments.schemas_gold],
                "discord_users",
            ),
        )

        canonical = most_recent_version(batch_df, fn.col("source_keys.id"))

        return merge_most_recent_version(
            delta_table=delta_table, updates=canonical, merge_keys=["source_keys.id"]
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--silver-catalog",
        dest="silver_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The catalog to use",
    )

    parser.add_argument(
        "--target-catalog",
        dest="target_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The catalog to use",
    )

    parser.add_argument(
        "--checkpoints-path-prefix",
        dest="checkpoints_path_prefix",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    texts_table = full_table_name(
        args.silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_texts_ao],
    )
    texts = spark.read.format("delta").table(texts_table)

    write_guilds(
        spark, args.silver_catalog, args.target_catalog, args.checkpoints_path_prefix
    ).awaitTermination()

    write_channels(
        spark, args.silver_catalog, args.target_catalog, args.checkpoints_path_prefix
    ).awaitTermination()

    write_threads(
        spark, args.silver_catalog, args.target_catalog, args.checkpoints_path_prefix
    ).awaitTermination()

    write_members(
        spark, args.silver_catalog, args.target_catalog, args.checkpoints_path_prefix
    ).awaitTermination()

    write_messages(
        spark,
        args.silver_catalog,
        args.target_catalog,
        texts,
        args.checkpoints_path_prefix,
    ).awaitTermination()

    write_users(
        spark, args.silver_catalog, args.target_catalog, args.checkpoints_path_prefix
    ).awaitTermination()


if __name__ == "__main__":
    main()
