import argparse
from typing import cast

from delta.tables import DeltaTable
from pyspark.sql import SparkSession, DataFrame
import pyspark.sql.functions as fn
from pyspark.sql.streaming import StreamingQuery

from trend_etl.functions.delta import (
    full_table_name,
    most_recent_version,
    earliest_version,
    merge_most_recent_version,
    merge_earliest_version,
)
import trend_etl.schema.args as arguments

program = "canonicalize_facebook"


def write_groups(
    spark: SparkSession,
    silver_catalog: str,
    target_catalog: str,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    silver_table = full_table_name(
        silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_facebook_groups_ao],
    )
    streaming = spark.readStream.format("delta").table(silver_table)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/"
        + arguments.schemas[arguments.schemas_gold]
        + "/"
        + "facebook_groups"
    )

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(
            spark,
            full_table_name(
                target_catalog,
                arguments.schemas[arguments.schemas_gold],
                "facebook_groups",
            ),
        )

        canonical = most_recent_version(batch_df, fn.col("source_keys.id"))

        return merge_most_recent_version(
            delta_table=delta_table, updates=canonical, merge_keys=["source_keys.id"]
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_posts(
    spark: SparkSession,
    silver_catalog: str,
    target_catalog: str,
    texts: DataFrame,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    silver_table = full_table_name(
        silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_facebook_posts_ao],
    )
    streaming = spark.readStream.format("delta").table(silver_table)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/"
        + arguments.schemas[arguments.schemas_gold]
        + "/"
        + "facebook_posts"
    )

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(
            spark,
            full_table_name(
                target_catalog,
                arguments.schemas[arguments.schemas_gold],
                "facebook_posts",
            ),
        )

        canonical = earliest_version(batch_df, fn.col("source_keys.id"))

        canonical_with_text = canonical.join(texts, "text_version", "left_outer")

        return merge_earliest_version(
            delta_table=delta_table,
            updates=canonical_with_text,
            merge_keys=["source_keys.id"],
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_comments(
    spark: SparkSession,
    silver_catalog: str,
    target_catalog: str,
    texts: DataFrame,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    silver_table = full_table_name(
        silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_facebook_comments_ao],
    )
    streaming = spark.readStream.format("delta").table(silver_table)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/"
        + arguments.schemas[arguments.schemas_gold]
        + "/"
        + "facebook_comments"
    )

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(
            spark,
            full_table_name(
                target_catalog,
                arguments.schemas[arguments.schemas_gold],
                "facebook_comments",
            ),
        )

        canonical = earliest_version(batch_df, fn.col("source_keys.id"))

        canonical_with_text = canonical.join(texts, "text_version", "left_outer")

        return merge_earliest_version(
            delta_table=delta_table,
            updates=canonical_with_text,
            merge_keys=["source_keys.id"],
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--silver-catalog",
        dest="silver_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The catalog to use",
    )

    parser.add_argument(
        "--target-catalog",
        dest="target_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The catalog to use",
    )

    parser.add_argument(
        "--checkpoints-path-prefix",
        dest="checkpoints_path_prefix",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    texts_table = full_table_name(
        args.silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_texts_ao],
    )
    texts = spark.read.format("delta").table(texts_table)

    write_groups(
        spark, args.silver_catalog, args.target_catalog, args.checkpoints_path_prefix
    ).awaitTermination()

    write_posts(
        spark,
        args.silver_catalog,
        args.target_catalog,
        texts,
        args.checkpoints_path_prefix,
    ).awaitTermination()

    write_comments(
        spark,
        args.silver_catalog,
        args.target_catalog,
        texts,
        args.checkpoints_path_prefix,
    ).awaitTermination()


if __name__ == "__main__":
    main()
