import argparse
from typing import cast

from delta.tables import DeltaTable
from pyspark.sql import SparkSession, DataFrame
import pyspark.sql.functions as fn
from pyspark.sql.streaming import StreamingQuery

from trend_etl.functions.delta import (
    full_table_name,
    most_recent_version,
    merge_most_recent_version,
    get_canonical_reddit,
    merge_canonical_reddit,
)
import trend_etl.schema.args as arguments

program = "canonicalize_reddit"


def write_subreddits(
    spark: SparkSession,
    silver_catalog: str,
    target_catalog: str,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    silver_table = full_table_name(
        silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_subreddits_ao],
    )
    streaming = spark.readStream.format("delta").table(silver_table)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/"
        + arguments.schemas[arguments.schemas_gold]
        + "/"
        + "subreddits"
    )

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(
            spark,
            full_table_name(
                target_catalog,
                arguments.schemas[arguments.schemas_gold],
                "subreddits",
            ),
        )

        canonical = most_recent_version(batch_df, fn.col("source_keys.name"))

        return merge_most_recent_version(
            delta_table=delta_table, updates=canonical, merge_keys=["source_keys.name"]
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_submissions(
    spark: SparkSession,
    silver_catalog: str,
    target_catalog: str,
    subreddit_idents: DataFrame,
    texts: DataFrame,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    silver_table = full_table_name(
        silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_reddit_submissions_ao],
    )
    streaming = spark.readStream.format("delta").table(silver_table)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/"
        + arguments.schemas[arguments.schemas_gold]
        + "/"
        + "reddit_submissions"
    )

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(
            spark,
            full_table_name(
                target_catalog,
                arguments.schemas[arguments.schemas_gold],
                "reddit_submissions",
            ),
        )

        # NOTE: filter out some submissions that may not link to the subreddits table
        # (this affects a small amount of data from pushshift)
        # See "Missing Subreddits" section https://wiki.trend.community/en/technology/data-lake
        filtered = batch_df.join(
            subreddit_idents,
            fn.col("source_keys.subreddit_id") == fn.col("id"),
            "left_semi",
        )

        canonical = get_canonical_reddit(filtered, "submission")

        canonical_with_text = canonical.join(texts, "text_version", "left_outer")

        return merge_canonical_reddit(
            delta_table=delta_table, updates=canonical_with_text, schema="submission"
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def write_replies(
    spark: SparkSession,
    silver_catalog: str,
    target_catalog: str,
    subreddit_idents: DataFrame,
    texts: DataFrame,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    silver_table = full_table_name(
        silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_reddit_replies_ao],
    )
    streaming = spark.readStream.format("delta").table(silver_table)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/"
        + arguments.schemas[arguments.schemas_gold]
        + "/"
        + "reddit_replies"
    )

    def merge_batch(batch_df: DataFrame, _batch_id: int) -> None:
        delta_table = DeltaTable.forName(
            spark,
            full_table_name(
                target_catalog,
                arguments.schemas[arguments.schemas_gold],
                "reddit_replies",
            ),
        )

        # NOTE: filter out some submissions that may not link to the subreddits table
        # (this affects a small amount of data from pushshift)
        # See "Missing Subreddits" section https://wiki.trend.community/en/technology/data-lake
        filtered = batch_df.join(
            subreddit_idents,
            fn.col("source_keys.subreddit_id") == fn.col("id"),
            "left_semi",
        )

        canonical = get_canonical_reddit(filtered, "reply")

        canonical_with_text = canonical.join(texts, "text_version", "left_outer")

        return merge_canonical_reddit(
            delta_table=delta_table, updates=canonical_with_text, schema="reply"
        )

    return (
        streaming.writeStream.format("delta")
        .foreachBatch(merge_batch)
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--silver-catalog",
        dest="silver_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The catalog to use",
    )

    parser.add_argument(
        "--target-catalog",
        dest="target_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The catalog to use",
    )

    parser.add_argument(
        "--checkpoints-path-prefix",
        dest="checkpoints_path_prefix",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    subreddit_idents_table = full_table_name(
        args.silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_subreddit_idents],
    )
    subreddit_idents = spark.read.format("delta").table(subreddit_idents_table)

    texts_table = full_table_name(
        args.silver_catalog,
        arguments.schemas[arguments.schemas_silver],
        arguments.tables[arguments.tables_texts_ao],
    )
    texts = spark.read.format("delta").table(texts_table)

    write_subreddits(
        spark, args.silver_catalog, args.target_catalog, args.checkpoints_path_prefix
    ).awaitTermination()

    write_submissions(
        spark,
        args.silver_catalog,
        args.target_catalog,
        subreddit_idents,
        texts,
        args.checkpoints_path_prefix,
    ).awaitTermination()

    write_replies(
        spark,
        args.silver_catalog,
        args.target_catalog,
        subreddit_idents,
        texts,
        args.checkpoints_path_prefix,
    ).awaitTermination()


if __name__ == "__main__":
    main()
