import argparse
from typing import cast

from delta.tables import DeltaTable
from pyspark.sql import SparkSession
from pyspark.sql.streaming import StreamingQuery

from trend_etl.functions.delta import (
    full_table_name,
    ao_merge_using_ingestion_key,
    ao_merge_using_text_key,
)
from trend_etl.transform.replies import (
    from_bronze_get_replies_df_ao,
    from_bronze_get_replies_text_df_ao,
)
import trend_etl.schema.args as arguments

program = "load_replies_ao"


def load_replies(
    spark: SparkSession,
    bronze_catalog: str,
    target_catalog: str,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    bronze_table = full_table_name(
        bronze_catalog,
        arguments.schemas[arguments.schemas_bronze],
        arguments.tables[arguments.tables_replies],
    )
    streaming = spark.readStream.format("delta").table(bronze_table)

    reply_idents = spark.read.format("delta").table(
        full_table_name(
            target_catalog,
            arguments.schemas[arguments.schemas_silver],
            arguments.tables[arguments.tables_reply_idents],
        )
    )

    prepared = from_bronze_get_replies_df_ao(streaming, reply_idents)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/reddit_replies_ao/"
        + arguments.tables[arguments.tables_reddit_replies_ao]
    )

    table = DeltaTable.forName(
        spark,
        full_table_name(
            target_catalog,
            arguments.schemas[arguments.schemas_silver],
            arguments.tables[arguments.tables_reddit_replies_ao],
        ),
    )

    return (
        prepared.writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: ao_merge_using_ingestion_key(table, batch_df)
        )
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def load_texts(
    spark: SparkSession,
    bronze_catalog: str,
    target_catalog: str,
    checkpoints_path_prefix: str,
) -> StreamingQuery:
    bronze_table = full_table_name(
        bronze_catalog,
        arguments.schemas[arguments.schemas_bronze],
        arguments.tables[arguments.tables_replies],
    )
    streaming = spark.readStream.format("delta").table(bronze_table)

    prepared = from_bronze_get_replies_text_df_ao(streaming)

    checkpoint_loc = (
        checkpoints_path_prefix
        + "/reddit_replies_ao/"
        + arguments.tables[arguments.tables_texts_ao]
    )

    table = DeltaTable.forName(
        spark,
        full_table_name(
            target_catalog,
            arguments.schemas[arguments.schemas_silver],
            arguments.tables[arguments.tables_texts_ao],
        ),
    )

    return (
        prepared.writeStream.format("delta")
        .foreachBatch(
            lambda batch_df, batch_id: ao_merge_using_text_key(table, batch_df)
        )
        .option("checkpointLocation", checkpoint_loc)
        .outputMode("append")
        .trigger(availableNow=True)
        .start()
    )


def main() -> None:
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{program}.")

    parser.add_argument(
        "--bronze-catalog",
        dest="bronze_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The unit catalog to use",
    )

    parser.add_argument(
        "--target-catalog",
        dest="target_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The unit catalog to use",
    )

    parser.add_argument(
        "--checkpoints-path-prefix",
        dest="checkpoints_path_prefix",
        required=True,
        help="The path to the folder (in S3) to store the checkpoints for this table",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(program).getOrCreate())

    load_replies(
        spark, args.bronze_catalog, args.target_catalog, args.checkpoints_path_prefix
    ).awaitTermination()

    load_texts(
        spark, args.bronze_catalog, args.target_catalog, args.checkpoints_path_prefix
    ).awaitTermination()


if __name__ == "__main__":
    main()
