"""
Validate that the annotations match the corresponding docs text field.
"""

import argparse
from typing import cast

from pyspark.sql import SparkSession
from pyspark.sql.functions import col, coalesce

from trend_etl.functions.delta import full_table_name
from trend_etl.functions.udfs import annotation_matches_text
import trend_etl.schema.args as arguments

PROGRAM = "validate_nlp_annotations"
OUTPUT_PATH = "dbfs:/FileStore/charlie/failed_annotation_matches"


def main() -> None:
    """
    Entry point for task
    """
    parser = argparse.ArgumentParser(description=f"Arguments for Trend ETL/{PROGRAM}.")

    parser.add_argument(
        "--target-catalog",
        dest="target_catalog",
        required=True,
        choices=arguments.catalogs,
        help="The unit catalog to use",
    )

    args = parser.parse_args()

    spark = cast(SparkSession, SparkSession.builder.appName(PROGRAM).getOrCreate())

    #### begin fetching table dataframes

    nlp_results = (
        spark.read.format("delta")
        .table(
            full_table_name(
                args.target_catalog,
                arguments.schemas[arguments.schemas_silver],
                arguments.tables[arguments.tables_historical_prod_nlp_results],
            )
        )
        .where(col("model_name").isin("csr", "csrl") & (col("model_version") == 1))
    )

    docs = spark.read.format("delta").table(
        full_table_name(
            args.target_catalog,
            arguments.schemas[arguments.schemas_silver],
            arguments.tables[arguments.tables_docs],
        )
    )

    posts = (
        spark.read.format("delta")
        .table(
            full_table_name(
                args.target_catalog,
                arguments.schemas[arguments.schemas_silver],
                arguments.tables[arguments.tables_facebook_posts],
            )
        )
        .selectExpr("doc_uuid", "source_keys.id as post_source_id")
    )

    comments = (
        spark.read.format("delta")
        .table(
            full_table_name(
                args.target_catalog,
                arguments.schemas[arguments.schemas_silver],
                arguments.tables[arguments.tables_facebook_comments],
            )
        )
        .selectExpr("doc_uuid", "source_keys.id as comment_source_id")
    )

    submissions = (
        spark.read.format("delta")
        .table(
            full_table_name(
                args.target_catalog,
                arguments.schemas[arguments.schemas_silver],
                arguments.tables[arguments.tables_reddit_submissions],
            )
        )
        .selectExpr("doc_uuid", "source_keys.name as submission_source_id")
    )

    replies = (
        spark.read.format("delta")
        .table(
            full_table_name(
                args.target_catalog,
                arguments.schemas[arguments.schemas_silver],
                arguments.tables[arguments.tables_reddit_replies],
            )
        )
        .selectExpr("doc_uuid", "source_keys.name as reply_source_id")
    )

    #### end

    docs_joined = nlp_results.join(
        docs.selectExpr("doc_uuid", "texts"), "doc_uuid", "left_outer"
    )

    failed_matches = docs_joined.where(
        ~annotation_matches_text(col("texts.text"), col("annotation"))
    )

    enriched_with_source_ids = (
        failed_matches.join(posts, "doc_uuid", "left_outer")
        .join(comments, "doc_uuid", "left_outer")
        .join(submissions, "doc_uuid", "left_outer")
        .join(replies, "doc_uuid", "left_outer")
        .withColumn(
            "source_id",
            coalesce(
                col("post_source_id"),
                col("comment_source_id"),
                col("submission_source_id"),
                col("reply_source_id"),
            ),
        )
    )

    (
        enriched_with_source_ids.repartition(1)
        .write.mode("overwrite")
        .option("overwriteSchema", "true")
        .format("delta")
        .save(OUTPUT_PATH)
    )
