import unittest
from unittest.mock import patch
from datetime import datetime

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.functions import col, lit

from trend_etl.test_utils import getspark, get_kafka_df, get_df
from trend_etl.functions.udfs import conv_timestamp
from trend_etl.transform.discord_stats import (
    total_journeys_number,
    journey_conversion_number,
    total_statements_number,
    active_users_for_threads,
    active_users_for_messages,
    active_users_for_reactions,
    total_registered_users_number,
    active_users_days_ago_number,
    active_audience_days_ago_number,
    retention_audience_number,
    days_ago_cond,
    greater_days_cond,
)
from trend_etl.transform.channels import (
    from_bronze_get_channels_df_ao,
    from_kafka_get_bronze_channels_df,
)
from trend_etl.transform.messages import (
    from_bronze_get_messages_df_ao,
    from_kafka_get_bronze_messages_df,
)
from trend_etl.transform.reactions import (
    from_bronze_get_reactions_df_ao,
    from_kafka_get_bronze_reactions_df,
)
from trend_etl.transform.members import (
    from_bronze_get_members_df_ao,
    from_bronze_get_users_df_ao,
    from_kafka_get_bronze_members_df,
)
import trend_etl.schema.args as arguments


class TestSilverTranforms(unittest.TestCase):
    spark: SparkSession
    channels: DataFrame
    threads: DataFrame
    messages: DataFrame
    reactions: DataFrame
    members: DataFrame
    users: DataFrame

    @classmethod
    def setUpClass(cls) -> None:
        cls.spark = getspark()

        message_idents = get_df(
            cls.spark,
            "doc_uuid string, id string",
            "test_data",
            "discord_stats",
            "message_idents.jsonl",
        )

        cls.channels = from_bronze_get_channels_df_ao(
            from_kafka_get_bronze_channels_df(
                get_kafka_df(cls.spark, "test_data", "channels.jsonl"), "channels"
            )
        )

        cls.threads = cls.channels.where(
            col("type").isin(arguments.discord_thread_types)
        )

        cls.messages = from_bronze_get_messages_df_ao(
            from_kafka_get_bronze_messages_df(
                get_kafka_df(cls.spark, "test_data", "discord_stats", "messages.jsonl"),
                "messages",
            ),
            message_idents,
        )

        cls.reactions = from_bronze_get_reactions_df_ao(
            from_kafka_get_bronze_reactions_df(
                get_kafka_df(cls.spark, "test_data", "reactions.jsonl"), "reactions"
            )
        )

        cls.members = from_bronze_get_members_df_ao(
            from_kafka_get_bronze_members_df(
                get_kafka_df(cls.spark, "test_data", "members.jsonl"), "members"
            )
        )

        cls.users = from_bronze_get_users_df_ao(
            from_kafka_get_bronze_members_df(
                get_kafka_df(cls.spark, "test_data", "members.jsonl"), "members"
            )
        )

    def test_total_journeys_number(self) -> None:
        self.assertEqual(
            1,
            total_journeys_number(self.channels, self.users, ["1215765928013201488"]),
            "Expected one non-trend employee to haved posted a journey",
        )

    def test_journeys_conversion_number(self) -> None:
        self.assertEqual(
            1,
            journey_conversion_number(
                self.channels, self.users, ["1215765928013201488"]
            ),
            "Expected one non-trend employee to haved posted a journey",
        )

    def test_total_statements_number(self) -> None:
        # 1 thread by non tredn employee
        # 22 messages by non trend employee
        self.assertEqual(
            23,
            total_statements_number(self.threads, self.messages),
            "Expected to have 20 threads + messages not authored by trend employees",
        )

    @patch("pyspark.sql.functions.current_timestamp")
    def test_active_users_for_threads(self, mock_current_timestamp) -> None:
        # Set the mock to return a fixed datetime
        fixed_current_timestamp = lit(datetime(2024, 4, 8))
        mock_current_timestamp.return_value = fixed_current_timestamp

        thread_created_column = conv_timestamp("thread_metadata.create_timestamp")
        threads_date_cond = days_ago_cond(thread_created_column, 5)
        active_users = active_users_for_threads(self.threads, threads_date_cond)

        self.assertEqual(
            active_users.count(),
            1,
            "Expects one thread created by a non-trend employee",
        )

    @patch("pyspark.sql.functions.current_timestamp")
    def test_active_users_for_messages(self, mock_current_timestamp) -> None:
        # Set the mock to return a fixed datetime
        fixed_current_timestamp = lit(datetime(2024, 4, 10))
        mock_current_timestamp.return_value = fixed_current_timestamp

        messages_date_cond = days_ago_cond(col("created_at"), 5)
        active_users = active_users_for_messages(self.messages, messages_date_cond)

        # NOTE: The 2 messages created on 2024-04-03 are older than the time filter
        self.assertEqual(
            active_users.count(),
            15,
            "Expects 15 messages created by non-trend employees",
        )

        messages_date_cond = greater_days_cond(col("created_at"), 5)
        active_users = active_users_for_messages(self.messages, messages_date_cond)
        self.assertEqual(
            active_users.count(),
            7,
            "Expects 7 messages created by non-trend employees",
        )

    @patch("pyspark.sql.functions.current_timestamp")
    def test_active_users_for_reactions(self, mock_current_timestamp) -> None:
        # Set the mock to return a fixed datetime
        fixed_current_timestamp = lit(datetime(2024, 4, 10))
        mock_current_timestamp.return_value = fixed_current_timestamp

        messages_date_cond = days_ago_cond(col("created_at"), 5)
        active_users = active_users_for_reactions(
            self.messages, self.reactions, self.users, messages_date_cond
        )

        self.assertEqual(
            active_users.count(),
            5,
            "Expects 5 reactions created by non-trend employees",
        )

    def test_total_registered_users_number(self) -> None:
        self.assertEqual(
            4,
            total_registered_users_number(self.users),
            "Expected 4 non-trend employees as members",
        )

    @patch("pyspark.sql.functions.current_timestamp")
    def test_active_users_days_ago_number(self, mock_current_timestamp) -> None:
        # Set the mock to return a fixed datetime
        fixed_current_timestamp = lit(datetime(2024, 4, 10))
        mock_current_timestamp.return_value = fixed_current_timestamp

        active_users = active_users_days_ago_number(self.threads, self.messages, 5, 1)

        self.assertEqual(
            active_users,
            2,
            "Expects 2 active users using message_count_threshold=1",
        )

        active_users = active_users_days_ago_number(self.threads, self.messages, 5, 4)

        self.assertEqual(
            active_users,
            1,
            "Expects 1 active users using message_count_threshold=4",
        )

    @patch("pyspark.sql.functions.current_timestamp")
    def test_active_audience_days_ago_number(self, mock_current_timestamp) -> None:
        # Set the mock to return a fixed datetime
        fixed_current_timestamp = lit(datetime(2024, 4, 10))
        mock_current_timestamp.return_value = fixed_current_timestamp

        active_audience = active_audience_days_ago_number(
            self.threads, self.messages, self.reactions, self.users, 5, 1, 1
        )

        self.assertEqual(
            active_audience,
            2,
            "Expects 2 active audience",
        )

    @patch("pyspark.sql.functions.current_timestamp")
    def test_retention_audience_number_number(self, mock_current_timestamp) -> None:
        # Set the mock to return a fixed datetime
        fixed_current_timestamp = lit(datetime(2024, 4, 10))
        mock_current_timestamp.return_value = fixed_current_timestamp

        retention_audience = retention_audience_number(
            self.threads, self.messages, self.reactions, self.users, 5
        )

        self.assertEqual(
            retention_audience,
            1,
            "Expects 1 for audience retention",
        )
