from pyspark.sql import DataFrame

import pyspark.sql.functions as fn
from pyspark.sql.types import StringType

import trend_etl.schema.fields as fields
from trend_etl.functions.misc import get_created_day
from trend_etl.functions.udfs import discord_snowflake_timestamp
from trend_etl.functions.parsing import parse_guilds
import trend_etl.functions.ingest_versioning as ingest
from trend_etl.schema.canonical import discord_guild_canonical


#
# arguments:
#   input_df: dataframe received from reading a kafka topic
#   target_topic: source topic from kafka
# return: a dataframe in the schema expected for bronze tables
def from_kafka_get_bronze_guilds_df(
    input_df: DataFrame, target_topic: str
) -> DataFrame:
    snowflake_id = discord_snowflake_timestamp(
        fn.get_json_object(fn.col(fields.kafka_value_field).cast(StringType()), "$.id")
    )

    return input_df.select(
        fn.col(fields.kafka_value_field).alias(fields.bronze_value_field),
        fn.lit(f"kafka/{target_topic}").alias(fields.bronze_provenance_field),
        fn.col(fields.kafka_timestamp_field).alias(fields.bronze_timestamp_field),
        get_created_day(snowflake_id).alias(fields.created_day_field),
    )


# arguments:
#   streaming_bronze: streaming dataframe from bronze table
#   guild_idents: guild uuid idents table
def from_bronze_get_guilds_df(
    streaming_bronze: DataFrame,
    guild_idents: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_guilds(fn.col(fields.bronze_value_field).cast(StringType())).alias(
            "value"
        ),
        fn.col(fields.bronze_timestamp_field),
    )

    flattened_df = parsed_df.select(
        fn.col("value.*"),
        fn.col(fields.bronze_timestamp_field),
    )

    # join with guild_idents
    result_df = (
        flattened_df.join(
            guild_idents,
            "id",
            "left_outer",
        )
        .withColumn("created_at", discord_snowflake_timestamp(fn.col("id")))
        .withColumn(fields.source_keys_field, fn.struct(fn.col("id").alias("id")))
        .drop("id")
    )

    return result_df


def from_bronze_get_guilds_df_ao(
    streaming_bronze: DataFrame,
    guild_idents: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_guilds(fn.col(fields.bronze_value_field).cast(StringType())).alias(
            "json"
        ),
        fn.col(fields.bronze_timestamp_field).alias(fields.ingested_on_field),
    )

    flattened_df = parsed_df.select(
        ingest.ingest_version(fn.col("json"), discord_guild_canonical).alias(
            fields.ingest_version_field
        ),
        fn.col("json.*"),
        fn.col(fields.ingested_on_field),
    )

    # join with guild_idents
    result_df = (
        flattened_df.join(
            guild_idents,
            "id",
            "left_outer",
        )
        .withColumn("created_at", discord_snowflake_timestamp(fn.col("id")))
        .withColumn(fields.source_keys_field, fn.struct(fn.col("id").alias("id")))
        .drop("id")
    )

    return result_df
