from pyspark.sql import DataFrame

import pyspark.sql.functions as fn
from pyspark.sql.types import StringType

import trend_etl.schema.fields as fields
from trend_etl.functions.misc import get_created_day, is_empty_or_white_space
from trend_etl.functions.udfs import conv_timestamp, discord_snowflake_timestamp
from trend_etl.functions.parsing import parse_messages
import trend_etl.functions.ingest_versioning as ingest
from trend_etl.schema.canonical import (
    discord_message_canonical,
    discord_text_canonical,
    discord_mention_user_canonical,
    discord_mention_channel_canonical,
)


#
# arguments:
#   input_df: dataframe received from reading a kafka topic
#   target_topic: source topic from kafka
# return: a dataframe in the schema expected for bronze tables
def from_kafka_get_bronze_messages_df(
    input_df: DataFrame, target_topic: str
) -> DataFrame:
    snowflake_id = discord_snowflake_timestamp(
        fn.get_json_object(fn.col(fields.kafka_value_field).cast(StringType()), "$.id")
    )

    return input_df.select(
        fn.col(fields.kafka_value_field).alias(fields.bronze_value_field),
        fn.lit(f"kafka/{target_topic}").alias(fields.bronze_provenance_field),
        fn.col(fields.kafka_timestamp_field).alias(fields.bronze_timestamp_field),
        get_created_day(snowflake_id).alias(fields.created_day_field),
    )


# arguments:
#   streaming_bronze: streaming dataframe from bronze table
#   guild_idents: guild uuid idents table
#   channel_idents: channel uuid idents table
#   message_idents: message uuid idents table
def from_bronze_get_messages_df(
    streaming_bronze: DataFrame,
    guild_idents: DataFrame,
    channel_idents: DataFrame,
    message_idents: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_messages(fn.col(fields.bronze_value_field).cast(StringType())).alias(
            "value"
        ),
        fn.col(fields.bronze_timestamp_field),
    )

    flattened_df = parsed_df.select(
        fn.col("value.*"),
        fn.col(fields.bronze_timestamp_field),
    )

    bot_author = fn.col("author.bot").isNotNull() & fn.col("author.bot")
    mentions_bot = fn.exists(
        "mentions", lambda mention: mention.bot.isNotNull() & mention.bot
    )
    filter_bot_messages = flattened_df.where(~bot_author & ~mentions_bot)

    result_df = (
        filter_bot_messages.withColumn("created_at", conv_timestamp("timestamp"))
        .drop("timestamp")
        .withColumn("edited_at", conv_timestamp("edited_timestamp"))
        .drop("edited_timestamp")
        .withColumn(
            fields.source_keys_field,
            fn.struct(
                fn.col("id").alias("id"),
                fn.col("channel_id").alias("channel_id"),
                fn.col("message_reference").alias("message_reference"),
            ),
        )  # join with channel_idents
        .join(
            channel_idents.withColumnRenamed("id", "channel_id").withColumnRenamed(
                "doc_uuid", "channel_uuid"
            ),
            "channel_id",
        )
        .drop("channel_id")
        # join with guild_idents
        .join(
            guild_idents.withColumnRenamed("id", "guild_id"),
            "guild_id",
        )
        .drop("guild_id")
        # join with message_idents
        .join(message_idents, "id")
        .drop("id")
        ###
        ### construct reference field from discord's message_reference field
        ###
        # join to get reference for: guild_id
        .join(
            guild_idents.withColumnRenamed("group_id", "__group_id").withColumnRenamed(
                "id", "guild_id"
            ),
            fn.col("message_reference.guild_id") == fn.col("guild_id"),
            "left_outer",
        )
        # join to get reference for: channel_id
        .join(
            channel_idents.withColumnRenamed(
                "doc_uuid", "__channel_uuid"
            ).withColumnRenamed("id", "channel_id"),
            fn.col("message_reference.channel_id") == fn.col("channel_id"),
            "left_outer",
        )
        # join to get reference for: message_id
        .join(
            message_idents.withColumnRenamed(
                "doc_uuid", "__message_uuid"
            ).withColumnRenamed("id", "message_id"),
            fn.col("message_reference.message_id") == fn.col("message_id"),
            "left_outer",
        )
        .withColumn(
            "reference",
            fn.struct(
                fn.col("message_reference.type").alias("type"),
                fn.col("__group_id").alias("group_id"),
                fn.col("__channel_uuid").alias("channel_uuid"),
                fn.col("__message_uuid").alias("message_uuid"),
            ),
        )
        .drop(
            "__group_id",
            "guild_id",
            "__channel_uuid",
            "channel_id",
            "__message_uuid",
            "message_id",
            "message_reference",
        )
    )

    return result_df


def from_bronze_get_messages_df_ao(
    streaming_bronze: DataFrame,
    message_idents: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_messages(fn.col(fields.bronze_value_field).cast(StringType())).alias(
            "json"
        ),
        fn.col(fields.bronze_timestamp_field).alias(fields.ingested_on_field),
    )

    flattened_df = parsed_df.select(
        ingest.ingest_version(fn.col("json"), discord_message_canonical).alias(
            fields.ingest_version_field
        ),
        fn.when(
            ~is_empty_or_white_space(fn.coalesce(fn.col("json.content"), fn.lit(""))),
            ingest.ingest_version(fn.col("json"), discord_text_canonical),
        ).alias(fields.text_version_field),
        fn.col("json.*"),
        fn.col(fields.ingested_on_field),
    )

    bot_author = fn.col("author.bot").isNotNull() & fn.col("author.bot")
    mentions_bot = fn.exists(
        "mentions", lambda mention: mention.bot.isNotNull() & mention.bot
    )

    result_df = (
        flattened_df.withColumn("created_at", conv_timestamp("timestamp"))
        .drop("timestamp")
        .withColumn("edited_at", conv_timestamp("edited_timestamp"))
        .drop("edited_timestamp")
        # join with message_idents
        .join(message_idents, "id")
        .withColumn(
            fields.source_keys_field,
            fn.struct(
                fn.col("id").alias("id"),
                fn.col("author.id").alias("user_id"),
                fn.col("channel_id").alias("channel_id"),
                fn.col("message_reference"),
            ),
        )
        .withColumn("bot_content_flag", bot_author | mentions_bot)
        .drop(
            "id",
            "channel_id",
            "guild_id",
            "content",
            "author",
            "mentions",
            "mention_channels",
            "message_reference",
        )
    )

    return result_df


def from_bronze_get_discord_text_df_ao(streaming_bronze: DataFrame) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_messages(fn.col(fields.bronze_value_field).cast(StringType())).alias(
            "json"
        )
    )

    result_df = parsed_df.select(
        ingest.ingest_version(fn.col("json"), discord_text_canonical).alias(
            fields.text_version_field
        ),
        fn.col("json.content").alias("text"),
    ).where(~is_empty_or_white_space(fn.coalesce(fn.col("json.content"), fn.lit(""))))

    return result_df


def from_bronze_get_discord_mention_user_df_ao(
    streaming_bronze: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_messages(fn.col(fields.bronze_value_field).cast(StringType())).alias(
            "json"
        ),
        fn.col(fields.bronze_timestamp_field).alias(fields.ingested_on_field),
    )

    result_df = (
        parsed_df.select(
            fn.col("json.id").alias("message_id"),
            fn.explode("json.mentions").alias("mentions"),
            fn.col(fields.ingested_on_field),
        )
        .withColumn("user_id", fn.col("mentions").getField("id"))
        .drop("mentions")
        .select(
            ingest.ingest_version(
                fn.struct(fn.col("message_id"), fn.col("user_id")),
                discord_mention_user_canonical,
            ).alias("ingest_version"),
            fn.struct(fn.col("message_id"), fn.col("user_id")).alias(
                fields.source_keys_field
            ),
            fn.col(fields.ingested_on_field),
        )
    )

    return result_df


def from_bronze_get_discord_mention_channel_df_ao(
    streaming_bronze: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_messages(fn.col(fields.bronze_value_field).cast(StringType())).alias(
            "json"
        ),
        fn.col(fields.bronze_timestamp_field).alias(fields.ingested_on_field),
    )

    result_df = (
        parsed_df.select(
            fn.col("json.id").alias("message_id"),
            fn.explode("json.mention_channels").alias("mention_channels"),
            fn.col(fields.ingested_on_field),
        )
        .withColumn("channel_id", fn.col("mention_channels").getField("id"))
        .drop("mention_channels")
        .select(
            ingest.ingest_version(
                fn.struct(fn.col("message_id"), fn.col("channel_id")),
                discord_mention_channel_canonical,
            ).alias("ingest_version"),
            fn.struct(fn.col("message_id"), fn.col("channel_id")).alias(
                fields.source_keys_field
            ),
            fn.col(fields.ingested_on_field),
        )
    )

    return result_df
