from pyspark.sql import DataFrame

import pyspark.sql.functions as fn
from pyspark.sql.types import StringType

import trend_etl.schema.fields as fields
from trend_etl.functions.misc import get_collected_day
from trend_etl.functions.parsing import parse_reactions
import trend_etl.functions.ingest_versioning as ingest
from trend_etl.schema.canonical import discord_reaction_canonical


#
# arguments:
#   input_df: dataframe received from reading a kafka topic
#   target_topic: source topic from kafka
# return: a dataframe in the schema expected for bronze tables
def from_kafka_get_bronze_reactions_df(
    input_df: DataFrame, target_topic: str
) -> DataFrame:
    return input_df.select(
        fn.col(fields.kafka_value_field).alias(fields.bronze_value_field),
        fn.lit(f"kafka/{target_topic}").alias(fields.bronze_provenance_field),
        fn.col(fields.kafka_timestamp_field).alias(fields.bronze_timestamp_field),
        get_collected_day(fn.col(fields.kafka_timestamp_field)).alias(
            fields.collected_day_field
        ),
    )


# arguments:
#   streaming_bronze: streaming dataframe from bronze table
#   message_idents: message uuid idents table
def from_bronze_get_reactions_df(
    streaming_bronze: DataFrame,
    message_idents: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_reactions(fn.col(fields.bronze_value_field).cast(StringType())).alias(
            "value"
        ),
        fn.col(fields.bronze_timestamp_field),
    )

    flattened_df = parsed_df.select(
        fn.col("value.*"),
        fn.col(fields.bronze_timestamp_field),
    )

    result_df = (
        flattened_df.withColumn(
            fields.source_keys_field,
            fn.struct(fn.col("message_id").alias("message_id")),
        )
        .join(message_idents.withColumnRenamed("id", "message_id"), "message_id")
        .drop("message_id")
    )

    return result_df


def from_bronze_get_reactions_df_ao(streaming_bronze: DataFrame) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_reactions(fn.col(fields.bronze_value_field).cast(StringType())).alias(
            "json"
        ),
        fn.col(fields.bronze_timestamp_field).alias(fields.ingested_on_field),
    )

    flattened_df = parsed_df.select(
        ingest.ingest_version(
            fn.col("json")
            .withField("user_id", fn.col("json.user.id"))
            .dropFields("user"),
            discord_reaction_canonical,
        ).alias(fields.ingest_version_field),
        fn.col("json.*"),
        fn.col(fields.ingested_on_field),
    )

    result_df = flattened_df.withColumn(
        fields.source_keys_field,
        fn.struct(fn.col("message_id"), fn.col("user.id").alias("user_id")),
    ).drop("message_id", "user")

    return result_df
