from pyspark.sql import DataFrame

import pyspark.sql.functions as fn
from pyspark.sql.types import (
    StringType,
    BinaryType,
    TimestampType,
    LongType,
    DoubleType,
)

import trend_etl.schema.fields as fields
import trend_etl.schema.args as arguments
from trend_etl.functions.misc import (
    jobid_2_timestamp,
    get_created_day,
    is_empty_or_white_space_reddit,
)
from trend_etl.functions.parsing import (
    parse_reddit_submissions,
    parse_pushshift_submissions,
)
from trend_etl.functions.udfs import is_scientific
import trend_etl.functions.ingest_versioning as ingest
from trend_etl.schema.canonical import (
    reddit_submission_canonical,
    reddit_submission_text_canonical,
)


def from_reddit_archive_get_bronze_submissions_df(archive_df: DataFrame) -> DataFrame:
    parsed = archive_df.withColumn(
        fields.archive_value_field,
        parse_reddit_submissions(fn.col(fields.archive_value_field)),
    )
    created_utc_field = fn.col(f"{fields.archive_value_field}.data.created_utc")

    return parsed.select(
        fn.to_json(fn.col(fields.archive_value_field))
        .cast(BinaryType())
        .alias(fields.bronze_value_field),
        fn.lit("archive/reddit/submissions").alias(fields.bronze_provenance_field),
        jobid_2_timestamp(fields.archive_job_id_field).alias(
            fields.bronze_timestamp_field
        ),
        get_created_day(created_utc_field.cast(TimestampType())).alias(
            fields.created_day_field
        ),
    )


def from_pushshift_archive_get_bronze_submissions_df(
    archive_df: DataFrame,
) -> DataFrame:
    parsed = archive_df.withColumn(
        fields.archive_value_field,
        parse_pushshift_submissions(fn.col(fields.archive_value_field)),
    )
    id_field = fn.col(f"{fields.archive_value_field}.id")
    created_utc_field = fn.col(f"{fields.archive_value_field}.created_utc")

    return (
        parsed.withColumn(
            fields.archive_value_field,
            fn.col(fields.archive_value_field).withField(
                "name",
                fn.format_string("t3_%s", id_field),
            ),
        )
        .withColumn(
            fields.archive_value_field,
            fn.col(fields.archive_value_field).withField(
                "created_utc",
                created_utc_field.cast(LongType()),
            ),
        )
        .select(
            fn.to_json(
                fn.struct(
                    fn.lit("t3").alias("kind"),
                    fn.col(fields.archive_value_field).alias("data"),
                    fn.lit("pushshift_archive").alias("note"),
                )
            )
            .cast(BinaryType())
            .alias(fields.bronze_value_field),
            fn.lit("archive/pushshift/submissions").alias(
                fields.bronze_provenance_field
            ),
            jobid_2_timestamp(fields.archive_job_id_field).alias(
                fields.bronze_timestamp_field
            ),
            get_created_day(created_utc_field.cast(TimestampType())).alias(
                fields.created_day_field
            ),
        )
    )


# arguments:
#   input_df: dataframe received from reading a kafka topic
#   target_topic: source topic from kafka
# return: a dataframe in the schema expected for bronze tables (for all reddit data kinds)
def from_kafka_get_bronze_submissions_df(input_df: DataFrame, target_topic: str):
    parsed = input_df.withColumn(
        fields.kafka_value_field,
        parse_reddit_submissions(fn.col(fields.kafka_value_field).cast(StringType())),
    )
    created_utc_field = fn.col(f"{fields.archive_value_field}.data.created_utc")

    return parsed.select(
        fn.to_json(fn.col(fields.kafka_value_field))
        .cast(BinaryType())
        .alias(fields.bronze_value_field),
        fn.lit(f"kafka/{target_topic}").alias(fields.bronze_provenance_field),
        fn.col(fields.kafka_timestamp_field).alias(fields.bronze_timestamp_field),
        get_created_day(created_utc_field.cast(TimestampType())).alias(
            fields.created_day_field
        ),
    )


# arguments:
#   streaming_bronze: streaming dataframe from bronze table
#   subreddit_idents: subreddit uuid idents table
#   submission_idents: submission uuid idents table
#   bot_authors: table of bot authors
# return: dataframe in the schema of `reddit_submissions`
def from_bronze_get_submissions_df(
    streaming_bronze: DataFrame,
    subreddit_idents: DataFrame,
    submission_idents: DataFrame,
    bot_authors: DataFrame,
) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_reddit_submissions(
            fn.col(fields.bronze_value_field).cast(StringType())
        ).alias("value"),
        fn.col(fields.bronze_timestamp_field),
    )
    flattened_df = parsed_df.select(
        fn.col("value.data.*"),
        fn.col(fields.bronze_timestamp_field),
    )

    # filter out bot submissions
    bot_filtered_df = flattened_df.join(
        bot_authors, fields.reddit_bot_users_author, "left_anti"
    )

    test_filtered_df = bot_filtered_df.where(
        fn.col("subreddit_id").isin(arguments.test_subreddits) == False
    )

    # NOTE: left outer joins (below) will cause any nulls to be
    #       detected as a downstream error.

    # join with subreddit_idents
    subreddits_df = test_filtered_df.join(
        subreddit_idents.withColumnRenamed(
            fields.idents_source_id_field, "subreddit_id"
        ),
        "subreddit_id",
        "inner",  # filter out submissions from subreddit_type = 'user' and like which we don't track
    ).drop("subreddit_id")

    # join with submission_idents
    submissions_df = (
        subreddits_df.join(
            submission_idents.withColumnRenamed(fields.idents_source_id_field, "name"),
            "name",
            "left_outer",
        )
        .withColumn(fields.source_keys_field, fn.struct(fn.col("name").alias("name")))
        .drop("name")
    )

    return (
        submissions_df.withColumn(
            fields.reddit_submissions_created_at,
            fn.when(
                is_scientific(fn.col("created_utc")),
                fn.col("created_utc").cast(DoubleType()).cast(LongType()),
            )
            .otherwise(fn.col("created_utc").cast(LongType()))
            .cast(TimestampType()),
        )
        .drop("created_utc")
        .withColumnRenamed("selftext", fields.reddit_submissions_text)
    )


def from_bronze_get_submissions_df_append(
    streaming_bronze: DataFrame,
    subreddit_idents: DataFrame,
    submission_idents: DataFrame,
) -> DataFrame:
    parsed = streaming_bronze.select(
        parse_reddit_submissions(
            fn.col(fields.bronze_value_field).cast(StringType())
        ).alias("value"),
        fn.col(fields.bronze_timestamp_field),
    )

    flattened = parsed.select(
        fn.col("value.data.*"),
        fn.col(fields.bronze_timestamp_field),
        fn.when(
            fn.coalesce(fn.col("value.note"), fn.lit("reddit")) == "pushshift_archive",
            "pushshift",
        )
        .otherwise("reddit")
        .alias("source_schema"),
    )

    # NOTE: `inner` join (below) will filter out submissions from
    # subreddit_type = 'user' and like which we don't track

    # join with subreddit_idents
    with_subreddits_idents = flattened.join(
        subreddit_idents.withColumnRenamed(
            fields.idents_source_id_field, "subreddit_id"
        ),
        "subreddit_id",
        "inner",
    ).drop("subreddit_id")

    # NOTE: `left_outer` join (below) will cause any nulls to be
    #       detected as a downstream error.

    # join with submission_idents
    with_submissions_idents = (
        with_subreddits_idents.join(
            submission_idents.withColumnRenamed(fields.idents_source_id_field, "name"),
            "name",
            "left_outer",
        )
        .withColumn(fields.source_keys_field, fn.struct(fn.col("name").alias("name")))
        .drop("name")
    )

    # parse created_utc epoch time
    # string that can be in scientific notation (float) or integer (long)
    created_parsed = with_submissions_idents.withColumn(
        fields.reddit_submissions_created_at,
        fn.when(
            is_scientific(fn.col("created_utc")),
            fn.col("created_utc").cast(DoubleType()).cast(LongType()),
        )
        .otherwise(fn.col("created_utc").cast(LongType()))
        .cast(TimestampType()),
    ).drop("created_utc")

    # rename text field
    return created_parsed.withColumnRenamed("selftext", fields.reddit_submissions_text)


def from_bronze_get_submissions_df_ao(
    streaming_bronze: DataFrame,
    submission_idents: DataFrame,
) -> DataFrame:
    parsed = streaming_bronze.select(
        parse_reddit_submissions(
            fn.col(fields.bronze_value_field).cast(StringType())
        ).alias("json"),
        fn.col(fields.bronze_timestamp_field).alias(fields.ingested_on_field),
    )

    flattened = parsed.select(
        ingest.ingest_version(
            fn.col("json").getField("data"), reddit_submission_canonical
        ).alias(fields.ingest_version_field),
        fn.when(
            ~is_empty_or_white_space_reddit(
                fn.coalesce(
                    fn.col("json").getField("data").getField("selftext"), fn.lit("")
                )
            ),
            ingest.ingest_version(
                fn.col("json").getField("data"), reddit_submission_text_canonical
            ),
        ).alias(fields.text_version_field),
        fn.col("json.data.*"),
        fn.col(fields.ingested_on_field),
        fn.when(
            fn.coalesce(fn.col("json.note"), fn.lit("reddit")) == "pushshift_archive",
            "pushshift",
        )
        .otherwise("reddit")
        .alias("source_schema"),
    )

    result_df = (
        flattened.join(
            submission_idents.withColumnRenamed(fields.idents_source_id_field, "name"),
            "name",
            "left_outer",
        )
        .withColumn(
            fields.source_keys_field,
            fn.struct(
                fn.col("name"),
                fn.col("author"),
                fn.col("author_fullname"),
                fn.col("subreddit_id"),
            ),
        )
        .withColumn(
            fields.reddit_submissions_created_at,
            fn.when(
                is_scientific(fn.col("created_utc")),
                fn.col("created_utc").cast(DoubleType()).cast(LongType()),
            )
            .otherwise(fn.col("created_utc").cast(LongType()))
            .cast(TimestampType()),
        )
        .drop(
            "created_utc",
            "name",
            "author",
            "author_fullname",
            "subreddit_id",
            "selftext",
        )
    )

    return result_df


def from_bronze_get_submissions_text_df_ao(streaming_bronze: DataFrame) -> DataFrame:
    parsed_df = streaming_bronze.select(
        parse_reddit_submissions(
            fn.col(fields.bronze_value_field).cast(StringType())
        ).alias("json")
    )

    result_df = parsed_df.select(
        ingest.ingest_version(
            fn.col("json").getField("data"), reddit_submission_text_canonical
        ).alias(fields.text_version_field),
        fn.col("json").getField("data").getField("selftext").alias("text"),
    ).where(
        ~is_empty_or_white_space_reddit(
            fn.coalesce(
                fn.col("json").getField("data").getField("selftext"), fn.lit("")
            )
        )
    )

    return result_df
